from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Optional, Union

from mloggers._log_levels import LogLevel


class Logger(ABC):
    """The abstract class for a logger."""

    @abstractmethod
    def log(
        self,
        message: Union[str, Dict[str, Any]],
        level: Optional[Union[LogLevel, str]] = None,
        *args: Any,
        **kwargs: Any,
    ):
        """
        Log a message.

        ### Parameters
        ----------
        `message`: the message to log.
        - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
        - If a dictionary, the message will be logged as a JSON string.
            - The dictionary must be JSON serializable.
            - You can provide None dictionary values to mean that the key is a header or title of the message.
        `level`: the level of the message (e.g., INFO, WARN, ERROR, DEBUG, etc.).
        - If None, no level will be printed.
        - If a string is provided, it will be colored in green (when colors are used) and uppercased; otherwise, the color will be the one associated with the `LogLevel` at time of registration.

        ### Raises
        ----------
        `TypeError`: if the message is not a string, a dictionary or does not implement `__str__()`.
        """

        if (
            not isinstance(message, dict)
            and not hasattr(message, "__str__")
            and not callable(getattr(message, "__str__"))
        ):
            raise TypeError(
                f"Expected message to be a string, a dictionary or to have implemented __str__(), but got {type(message)}."
            )

    def _call_impl(self, *args, **kwargs):
        return self.log(*args, **kwargs)

    __call__: Callable[..., Any] = _call_impl

    def info(
        self,
        message: Union[str, Dict[str, Any]],
        *args: Any,
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with level=LogLevel.INFO.

        ### Parameters
        ----------
        `message`: the message to log.
        - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
        - If a dictionary, the message will be logged as a JSON string.
            - The dictionary must be JSON serializable.
            - You can provide None dictionary values to mean that the key is a header or title of the message.
        """

        self.log(message, LogLevel.INFO, *args, **kwargs)

    def warn(
        self,
        message: Union[str, Dict[str, Any]],
        *args: Any,
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with level=LogLevel.WARN.

        ### Parameters
        ----------
        `message`: the message to log.
        - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
        - If a dictionary, the message will be logged as a JSON string.
            - The dictionary must be JSON serializable.
            - You can provide None dictionary values to mean that the key is a header or title of the message.
        """

        self.log(message, LogLevel.WARN, *args, **kwargs)

    def error(
        self,
        message: Union[str, Dict[str, Any]],
        *args: Any,
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with level=LogLevel.ERROR.

        ### Parameters
        ----------
        `message`: the message to log.
        - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
        - If a dictionary, the message will be logged as a JSON string.
            - The dictionary must be JSON serializable.
            - You can provide None dictionary values to mean that the key is a header or title of the message.
        """

        self.log(message, LogLevel.ERROR, *args, **kwargs)

    def debug(
        self,
        message: Union[str, Dict[str, Any]],
        *args: Any,
        **kwargs: Any,
    ):
        """
        Wrapper for calling `log` with level=LogLevel.DEBUG.

        ### Parameters
        ----------
        `message`: the message to log.
        - If a stringifiable object (implements `__str__()`), the message will be logged as-is.
        - If a dictionary, the message will be logged as a JSON string.
            - The dictionary must be JSON serializable.
            - You can provide None dictionary values to mean that the key is a header or title of the message.
        """

        self.log(message, LogLevel.DEBUG, *args, **kwargs)



import json
import os
from datetime import datetime
from typing import Any, Dict, Optional, Union

import numpy as np
from termcolor import colored

from mloggers._log_levels import LogLevel
from mloggers.logger import Logger


class FileLogger(Logger):
    """Logs to a file."""

    def __init__(self, file_path: str):
        """
        Initializes a file logger.

        ### Parameters
        ----------
        `file_path`: the path to the file to log to.
        - The file will be created if it does not exist. If it does, the logs will be appended to it.
        """

        # Create the file if it does not exist
        if not os.path.exists(file_path):
            with open(file_path, "w") as file:
                file.write("[]")

        print(f'{colored("[INFO]", "cyan")} Logging to file {file_path}')

        self._file_path = file_path

    def log(
        self,
        message: Union[str, Dict[str, Any]],
        level: Optional[Union[LogLevel, str]] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super(FileLogger, self).log(message, level, *args, **kwargs)

        # Convert numpy's ndarrays to lists so that they are JSON serializable
        if isinstance(message, dict):
            for key, value in message.items():
                if isinstance(value, np.ndarray):
                    message[key] = value.tolist()
        elif hasattr(message, "__str__") and callable(getattr(message, "__str__")):
            message = str(message)

        try:
            with open(self._file_path) as file:
                try:
                    logs = json.load(file)
                except json.decoder.JSONDecodeError:
                    print(
                        f'{colored("[ERROR]", "red")} Could not read the log file. Logs will not be saved.'
                    )

            log = {
                "timestamp": datetime.now().strftime("%d/%m/%Y %H:%M:%S"),
                "message": message,
            }
            if level is not None:
                log["level"] = level.name if isinstance(level, LogLevel) else str(level).upper()
            logs.append(log)

            with open(self._file_path, "w") as file:
                file.seek(0)
                json.dump(logs, file, indent=4)

        except Exception as e:
            print(
                f'{colored("[ERROR]", "red")} Exception thrown while logging to a file: {e}'
            )
